/*############################################################################
#                                                                            #
#   		                  DHCPNCAg								 #
# Qlj                                                                     #
# http://www.wiznet.co.kr/UpLoad_Files/ReferenceFiles/W5100_Appnote_DHCP.zip #
#                                                                            #
############################################################################*/
#define __DHCPC_C__

#include <avr/io.h>
#include <string.h>
#include <stdio.h>
#include <avr/pgmspace.h>
#include "net.h"

/* DHCP NCAgݒ	*/
#define DHCP_CLIENT_PORT	68			/* DHCPNCAg|[gԍ(UDP)	*/
#define	DHCP_SERVER_PORT	67			/* DHCPT[o[|[gԍ(UDP)		*/

#define	DHCP_TIMEOUT		200			/* ^CAEg(10msecP)			*/
#define	DHCP_RETRY			5			/* gC						*/	

/* DHCP message OP code 	*/
#define DHCP_BOOTREQUEST	1
#define DHCP_BOOTREPLY		2

/* DHCP message type */
#define	DHCP_DISCOVER	1
#define DHCP_OFFER		2
#define	DHCP_REQUEST	3
#define	DHCP_DECLINE	4
#define	DHCP_ACK		5
#define DHCP_NAK		6
#define	DHCP_RELEASE	7
#define DHCP_INFORM		8

#define MAGIC_COOKIE 0x63825363

/* bZ[W`\	(548bytes)	*/
typedef struct DHCP_MSG_t {
	uint8_t	op;
	uint8_t	htype;
	uint8_t	hlen;
	uint8_t	hops;
	uint32_t xid;
	uint16_t secs;
	uint16_t flags;
	uint8_t	ciaddr[4];
	uint8_t	yiaddr[4];
	uint8_t	siaddr[4];
	uint8_t	giaddr[4];
	uint8_t	chaddr[16];
	uint8_t	sname[64];
	uint8_t	file[128];
	uint8_t	OPT[312];
} DHCP_MSG;

/* [J֐vg^Cv錾		*/
static uint8_t dhcpClient_recvMessage( void );
static void dhcpClient_sendDISCOVER( void );
static void dhcpClient_sendREQUEST( void );
static uint8_t dhcpClient_openSocket( void );
static void setW5100_parameter( void );
static void reset_ip( void );
static void update_ip( void );

/*****************************************************************************
*							DHCPNCAg
* Fvoid
* ߒlFvoid
*****************************************************************************/
void dhcpc_init( void )
{
	/* netInfo \̏			*/
	if ( netInfo.dhcpMode == IP_DYNAMIC ) {
		netInfo.ip = 0L;
		netInfo.subnet = 0L;
		netInfo.gateway = 0L;
		netInfo.dns = 0L;
	}

	/* DHCP Ǘ\̏		*/
	memset( &dhcpInfo ,0x00 ,sizeof(DHCP_INFO) );
	dhcpInfo.xid = 0x19720517;
	if ( netInfo.dhcpMode == IP_DYNAMIC ) {
		dhcpInfo.status = DHCP_DISCOVER;
	} else {
		dhcpInfo.status = DHCP_STS_DISABLE;
	}

	/* W5100 p[^ݒ			*/
	setW5100_parameter();
}

/*****************************************************************************
*							DHCPNCAg^XN
* Fvoid
* ߒlFvoid
*****************************************************************************/
void dhcpc_task( void )
{
	if ( dhcpInfo.status == DHCP_STS_DISABLE ) {	/* DHCP?			*/
		return;
	}

	/* [XԐ؂`FbN(10؂烊Zbg)	*/
	if ( dhcpInfo.status >= DHCP_STS_LEASED ) {		/* [Xȏ̏?	*/
		if ( dhcpInfo.leaseTime != 0xffffffff ) {	/* L̃[XH		*/
			if ( dhcpInfo.ipTimer < 600L ) {		/* 10؂?			*/
				reset_ip();							/* DHCPZbg			*/
			}
		}
	}

	/* bZ[WM			*/
	uint8_t msgType = dhcpClient_recvMessage();

	/* DISCOVER 				*/
	if ( dhcpInfo.status == DHCP_STS_DISCOVER ) {
		if ( msgType == DHCP_OFFER ) {			/* OFFERM?			*/
			dhcpInfo.retry = DHCP_RETRY;
			dhcpInfo.timeOut = 0;
			dhcpInfo.status = DHCP_STS_REQUEST;

		} else if ( dhcpInfo.timeOut == 0 ) {	/* DISCOVER M		*/
			dhcpClient_sendDISCOVER();
			dhcpInfo.timeOut = DHCP_TIMEOUT;
		}

	/* REQUEST 					*/
	} else if ( dhcpInfo.status == DHCP_STS_REQUEST ) {
		if ( msgType == DHCP_ACK ) {			/* ACKMH			*/
			udp_close( DHCP_SOCK );				/* DHCP\PbgN[Y	*/
			if ( dhcpInfo.oldIP != dhcpInfo.getIP ) {
				update_ip();					/* IPAhXXV		*/
			}
			dhcpInfo.status = DHCP_STS_LEASED;

		} else if ( msgType == DHCP_NAK ) {		/* NAKMH			*/
			reset_ip();

		} else if ( dhcpInfo.timeOut == 0 ) {	/* REQUEST M			*/
			if ( dhcpInfo.retry ) {
				dhcpClient_sendREQUEST();
				dhcpInfo.timeOut = DHCP_TIMEOUT;
				dhcpInfo.retry--;
			} else {							/* gCI[o[H	*/
				reset_ip();
			}
		}

	/* LEASED 					*/
	} else if ( dhcpInfo.status == DHCP_STS_LEASED ) {
		if ( dhcpInfo.ipTimer <= ( dhcpInfo.leaseTime / 3L ) ) {		/* ṽ^C~OH	*/
			dhcpInfo.oldIP = dhcpInfo.getIP;
			dhcpInfo.xid++;
			dhcpInfo.timeOut = 0;
			dhcpInfo.status = DHCP_STS_REREQUEST;
		}

	/* REREQUEST 				*/
	} else if ( dhcpInfo.status == DHCP_STS_REREQUEST ) {
		if ( msgType == DHCP_ACK ) {			/* ACKMH			*/
			udp_close( DHCP_SOCK );				/* DHCP\PbgN[Y	*/
			if ( dhcpInfo.oldIP != dhcpInfo.getIP ) {
				update_ip();					/* IPAhXXV		*/
			}
			dhcpInfo.status = DHCP_STS_LEASED;

		} else if ( msgType == DHCP_NAK ) {		/* NAKMH			*/
			dhcpInfo.status = DHCP_STS_DISCOVER;
			dhcpInfo.timeOut = DHCP_TIMEOUT;

		} else if ( dhcpInfo.timeOut == 0 ) {	/* REQUEST M			*/
			if ( dhcpInfo.retry ) {
				dhcpClient_sendREQUEST();
				dhcpInfo.retry--;
			} else {							/* gCI[o[H	*/
				dhcpInfo.status = DHCP_STS_DISCOVER;
			}
			dhcpInfo.timeOut = DHCP_TIMEOUT;
		}
	}
}

/*****************************************************************************
*							bZ[WM
* Fvoid
* ߒlFuint8_t bZ[W^Cv
*****************************************************************************/
static uint8_t dhcpClient_recvMessage( void )
{
	DHCP_MSG msg ,*pMsg = &msg;
	uint32_t s_ip = 0L;
	uint16_t s_port = 0;
	uint16_t len;
	uint8_t *p;
	uint8_t *e;
	uint8_t type = 0;
	uint8_t opt_len = 0;
	uint8_t c;

	/* \Pbgԃ`FbN			*/
	if ( ( wiz_getSn_SR( DHCP_SOCK ) != SOCK_UDP ) || ( wiz_getSn_SPORT( DHCP_SOCK ) != DHCP_CLIENT_PORT ) ) {
		return ( 0 );
	}

	/* UDP\Pbgf[^擾	*/
	len = udp_recvfrom( DHCP_SOCK ,pMsg ,sizeof(DHCP_MSG) ,&s_ip ,&s_port );
	if ( ( len < 240 ) || ( len > sizeof(DHCP_MSG) ) ) {
		return ( 0 ); 
	}

	/* DHCPbZ[W`FbN		*/
	if ( ( msg.op != DHCP_BOOTREPLY ) || 
	     ( s_port != DHCP_SERVER_PORT ) ||
		 ( memcmp( msg.chaddr ,netInfo.mac ,6 ) != 0 ) ||
		 ( msg.xid != swapByteorder32( dhcpInfo.xid ) ) ) {
		return ( 0 );
	}

	/* DHCPT[o[ƒʐMĂꍇ̑IP`FbN	*/
	if ( dhcpInfo.dhcp_srever_ip != 0 ) {
		if ( dhcpInfo.dhcp_srever_ip != s_ip ) {
			return ( 0 );
		}
	}

	/* IvVf[^ւ̃|C^vZ		*/
	p = (uint8_t *)pMsg + 240;
	e = p + ( len - 240 );

	/* IvVf[^					*/
	while ( p < e ) {
		c = *p++;
		if ( c == 255 ) {										/* endOption ?							*/
			return ( type );

		} else if ( c == 53 ) {									/* dhcpMessageType ?					*/
			opt_len = *p++;
			type = *p;

		} else if ( c == 1 ) {									/* subnetMask ?							*/
			opt_len = *p++;
			dhcpInfo.getSUBNET = *(uint32_t *)p;

		} else if ( c == 3 ) {									/* routersOnSubnet ?					*/
			opt_len = *p++;
			dhcpInfo.getGATEWAY = *(uint32_t *)p;

		} else if ( c == 6 ) {									/* dns ?								*/
			opt_len = *p++;
			dhcpInfo.getDNS = *(uint32_t *)p;

		} else if ( c == 51 ) {									/* dhcpIPaddrLeaseTime ?				*/
			opt_len = *p++;
			dhcpInfo.leaseTime = swapByteorder32( *(uint32_t *)p );
			dhcpInfo.ipTimer = dhcpInfo.leaseTime;

		} else if ( c == 54 ) {									/* dhcpServerIdentifier ?				*/
			opt_len = *p++;

			if ( ( dhcpInfo.dhcp_srever_ip == 0 ) || ( dhcpInfo.dhcp_srever_ip == s_ip ) ) {
				dhcpInfo.dhcp_srever_ip = s_ip;					/* ĂꂽDHCPT[o[IPޔ	*/
				dhcpInfo.getIP = *(uint32_t *)msg.yiaddr;		/* ʒmꂽIPAhXޔ			*/
			}

		} else {
			opt_len = *p++;
		}

		p += opt_len;
	}

	return ( 0 );
}

/*****************************************************************************
*						DHCPT[o[ DISCOVER M
* Fvoid
* ߒlFvoid
*****************************************************************************/
static void dhcpClient_sendDISCOVER( void )
{
	uint16_t len;
	uint32_t ip = 0xffffffff;
	DHCP_MSG msg;
	uint8_t *p = msg.OPT;

	if ( dhcpClient_openSocket() != 0 ) {			/* \PbgI[v			*/
		return;
	}

	memset( &msg ,0x00 ,sizeof(DHCP_MSG) );
	msg.op = DHCP_BOOTREQUEST;
	msg.htype = 1;									/* HTYPE10MB				*/
	msg.hlen = 6;									/* HLENETHERNET				*/
	msg.hops = 0;									/* HOPS						*/
	msg.xid = swapByteorder32( dhcpInfo.xid );		/* XID						*/
	msg.secs = 0;									/* SECS (0)					*/
	msg.flags = swapByteorder16( 0x8000 );			/* FLAGSBROADCAST			*/
	memcpy( msg.chaddr ,netInfo.mac ,6 );			/* chaddr = MAC				*/

	*(uint32_t *)p = swapByteorder32( MAGIC_COOKIE );	/* MAGIC_COOKIE			*/
	p += 4;

	*p++ = 53;										/* dhcpMessageType			*/
	*p++ = 0x01;
	*p++ = DHCP_DISCOVER;

	*p++ = 61;										/* dhcpClientIdentifier		*/
	*p++ = 0x07;
	*p++ = 0x01;
	memcpy( p ,netInfo.mac ,6 );
	p += 6;

	*p++ = 12;										/* hostName					*/
	len = strlen( netInfo.hostname );
	*p++ = ( len + 13 );
	memcpy( p ,netInfo.hostname ,len );
	p += len;
	sprintf_P( (char *)p ,PSTR("-%02X%02X%02X%02X%02X%02X") ,netInfo.mac[0] ,netInfo.mac[1] ,netInfo.mac[2] ,netInfo.mac[3] ,netInfo.mac[4] ,netInfo.mac[5] );
	p += 13;

	*p++ = 55;										/* dhcpParamRequest			*/
	*p++ = 6;
	*p++ = 1;										/* subnetMask				*/
	*p++ = 3;										/* routersOnSubnet			*/
	*p++ = 6;										/* dns						*/
	*p++ = 15;										/* domainName				*/
	*p++ = 58;										/* dhcpT1value				*/
	*p++ = 59;										/* dhcpT2value				*/
	*p++ = 255;										/* endOption				*/

	/* u[hLXgM			*/
	udp_sendto( DHCP_SOCK ,&msg ,sizeof(DHCP_MSG) ,ip ,DHCP_SERVER_PORT );
}

/*****************************************************************************
*					DHCPT[o[ REQUEST M
* Fvoid
* ߒlFvoid
*****************************************************************************/
static void dhcpClient_sendREQUEST( void )
{
	uint16_t len;
	uint32_t ip;

	DHCP_MSG msg;
	uint8_t *p = msg.OPT;

	if ( dhcpClient_openSocket() != 0 ) {			/* \PbgI[v			*/
		return;
	}

	memset( &msg ,0x00 ,sizeof(DHCP_MSG) );
	msg.op = DHCP_BOOTREQUEST;
	msg.htype = 1;									/* HTYPE10MB				*/
	msg.hlen = 6;									/* HLENETHERNET				*/
	msg.hops = 0;									/* HOPS						*/
	msg.xid = swapByteorder32( dhcpInfo.xid );		/* XID						*/
	msg.secs = 0;									/* SECS (0)					*/

	if ( dhcpInfo.status < DHCP_STS_LEASED ) {
		msg.flags = swapByteorder16( 0x8000 );		/* DHCP_FLAGSBROADCAST		*/
	} else {
		*(uint32_t *)msg.ciaddr = dhcpInfo.dhcp_srever_ip;
	}

	memcpy( msg.chaddr ,netInfo.mac ,6 );			/* chaddr = MAC				*/

	*(uint32_t *)p = swapByteorder32( MAGIC_COOKIE );	/* MAGIC_COOKIE			*/
	p += 4;

	*p++ = 53;										/* dhcpMessageType			*/
	*p++ = 1;
	*p++ = DHCP_REQUEST;

	*p++ = 61;										/* dhcpClientIdentifier		*/
	*p++ = 7;
	*p++ = 1;
	memcpy( p ,netInfo.mac ,6 );
	p += 6;

	if ( dhcpInfo.status < DHCP_STS_LEASED ) {
		*p++ = 50;									/* dhcpRequestedIPaddr		*/
		*p++ = 4;
		*(uint32_t *)p = dhcpInfo.getIP;
		p += 4;

		*p++ = 54;									/* dhcpServerIdentifier		*/
		*p++ = 4;
		*(uint32_t *)p = dhcpInfo.dhcp_srever_ip;
		p += 4;
	}

	*p++ = 12;										/* hostName					*/
	len = strlen( netInfo.hostname );
	*p++ = ( len + 13 );
	memcpy( p ,netInfo.hostname ,len );
	p += len;
	sprintf_P( (char *)p ,PSTR("-%02X%02X%02X%02X%02X%02X") ,netInfo.mac[0] ,netInfo.mac[1] ,netInfo.mac[2] ,netInfo.mac[3] ,netInfo.mac[4] ,netInfo.mac[5] );
	p += 13;

	*p++ = 55;										/* dhcpParamRequest			*/
	*p++ = 8;
	*p++ = 1;										/* subnetMask				*/
	*p++ = 3;										/* routersOnSubnet			*/
	*p++ = 6;										/* dns						*/
	*p++ = 15;										/* domainName				*/
	*p++ = 58;										/* dhcpT1value				*/
	*p++ = 59;										/* dhcpT2value				*/
	*p++ = 31;										/* performRouterDiscovery	*/
	*p++ = 33;										/* staticRoute				*/
	*p++ = 255;										/* endOption				*/

	/* 󋵂ɉăjLXg^u[hLXgI	*/
	if ( dhcpInfo.status == DHCP_STS_REREQUEST ) {	/* [Xv̓jLXg	*/
		ip = dhcpInfo.dhcp_srever_ip;
	} else {										/* ʂ̃NGXg̓u[hLXg	*/
		ip = 0xffffffff;
	}

	/* UDPM					*/
	udp_sendto( DHCP_SOCK ,&msg ,sizeof(DHCP_MSG) ,ip ,DHCP_SERVER_PORT );
}

/*****************************************************************************
*						DHCPNCAg\PbgI[v
* Fvoid
* ߒlFuint8_t 0:OK !0:NG
*****************************************************************************/
static uint8_t dhcpClient_openSocket( void )
{
	if ( ( wiz_getSn_SR( DHCP_SOCK ) != SOCK_UDP ) || ( wiz_getSn_SPORT( DHCP_SOCK ) != DHCP_CLIENT_PORT ) ) {
		if ( udp_open( DHCP_SOCK ,DHCP_CLIENT_PORT ) != SOCK_UDP ) {
			return ( 1 );
		}
	}
	return ( 0 );
}

/*****************************************************************************
*						W5100 lbg[Np[^ݒ
* Fvoid
* ߒlFvoid
*****************************************************************************/
static void setW5100_parameter( void )
{
	wiz_setSHAR( netInfo.mac );
	wiz_setSIPR( (uint8_t *)&netInfo.ip );
	wiz_setSUBR( (uint8_t *)&netInfo.subnet );
	wiz_setGAR( (uint8_t *)&netInfo.gateway );
}

/*****************************************************************************
*					ݐݒ肳ĂIPAhXj
* Fvoid
* ߒlFvoid
*****************************************************************************/
static void reset_ip( void )
{
	uint32_t xid;

	/* DHCP Ǘ\̃NA			*/
	xid = dhcpInfo.xid + 1;
	if ( xid == 0xffffffff ) {
		xid = 0x12345678;
	}

	memset( &dhcpInfo ,0x00 ,sizeof(DHCP_INFO) );
	dhcpInfo.status = DHCP_DISCOVER;
	dhcpInfo.xid = xid;

	netInfo.ip = 0L;
	netInfo.subnet = 0L;
	netInfo.gateway = 0L;
	netInfo.dns = 0L;

	setW5100_parameter();
}

/*****************************************************************************
*								IPAhXݒ
* Fvoid
* ߒlFvoid
*****************************************************************************/
static void update_ip( void )
{
	netInfo.ip = dhcpInfo.getIP;
	netInfo.subnet = dhcpInfo.getSUBNET;
	netInfo.gateway = dhcpInfo.getGATEWAY;
	netInfo.dns = dhcpInfo.getDNS;
	setW5100_parameter();
}

