/*###########################################################################
#                                                                           #
#   		                  DNSNCAg								#
# http://www.wiznet.co.kr/UpLoad_Files/ReferenceFiles/W5100_Appnote_DNS.zip #
#                                                                           #
###########################################################################*/
#define __DNSC_C__

#include <string.h>
#include <avr/io.h>
#include <util/delay.h>
#include "net.h"

#define DEBUG 0					/* 0:ʏ !0:Wo͂ɃfobNo				*/

/* CLASS 						*/
#define	CLASS_IN	1			/* the Internet 										*/
#define CLASS_CS	2			/* the CSNET class (Obsolete - used only for examples in some obsolete RFCs) */
#define CLASS_CH	3			/* the CHAOS class										*/
#define CLASS_HS	4			/* Hesiod [Dyer 87]										*/

/* QCLASS 						*/
#define QCLASS_ANY	255 		/* any class 											*/

/* 	QTYPE values																				*/
/*	TYPE fields are used in resource records.  Note that these types are a subset of QTYPEs.	*/

/* TYPE 						*/
#define	TYPE_A		1 			/* The ARPA Internet 									*/
#define	TYPE_NS		2 			/* an authoritative name server 						*/
#define	TYPE_MD		3 			/* a mail destination (Obsolete - use MX) 				*/
#define	TYPE_MF		4 			/* a mail forwarder (Obsolete - use MX) 				*/
#define	TYPE_CNAME	5 			/* the canonical name for an alias 						*/
#define	TYPE_SOA	6 			/* marks the start of a zone of authority				*/
#define	TYPE_MB		7			/* a mailbox domain name (EXPERIMENTAL) 				*/
#define	TYPE_MG		8		 	/* a mail group member (EXPERIMENTAL) 					*/
#define	TYPE_MR		9 			/* a mail rename domain name (EXPERIMENTAL)				*/
#define	TYPE_NULL	10 			/* a null RR (EXPERIMENTAL) 							*/
#define	TYPE_WKS	11 			/* a well known service description 					*/
#define	TYPE_PTR	12 			/* a domain name pointer 								*/
#define	TYPE_HINFO	13 			/* host information 									*/
#define	TYPE_MINFO	14 			/* mailbox or mail list information 					*/
#define	TYPE_MX		15 			/* mail exchange 										*/
#define	TYPE_TXT	16 			/* text strings 										*/

/* QTYPE 						*/
#define	QTYPE_AXFR		252		/* A request for a transfer of an entire zone 			*/
#define	QTYPE_MAILB		253 	/* A request for mailbox-related records (MB, MG or MR) */
#define	QTYPE_MAILA		254 	/* A request for mail agent RRs (Obsolete - see MX) 	*/
#define	QTYPE_TYPE_ALL	255 	/* A request for all records 							*/

#define	INITRTT		2000L		/* Initial smoothed response time 						*/
#define	MAXCNAME	10			/* Maximum amount of cname recursion					*/


/* Round trip timing parameters */
#define	AGAIN	8				/* Average RTT gain = 1/8			*/
#define	LAGAIN	3				/* Log2(AGAIN)						*/
#define	DGAIN	4				/* Mean deviation gain = 1/4		*/
#define	LDGAIN	2				/* log2(DGAIN) 						*/

#define	IPPORT_DOMAIN	53		/* DNST[o[|[gԍ			*/

/*																	*/
/* A four bit field that specifies kind of query in this			*/
/* message.  This value is set by the originator of a query			*/
/* and copied into the response.  The values are:					*/
/*																	*/
#define OP_QUERY		0		/* a standard query (QUERY) 		*/
#define OP_IQUREY		1		/* an inverse query (IQUERY) 		*/
#define OP_STATUS		2		/* a server status request (STATUS)	*/

/* A one bit field that specifies whether this message is a query (0), or a response (1). */
#define	QR_QUERY		0	
#define	QR_RESPONSE		1

/*																				*/
/* Response code - this 4 bit field is set as part of							*/
/*                 responses.  The values have the following interpretation:	*/
/*																				*/
#define	RC_NO_ERROR		0	/* No error condition */
#define	RC_FORMAT_ERROR	1	/* Format error - The name server was unable to interpret the query. */
#define	RC_SERVER_FAIL	2	/* Server failure - The name server was unable to process this query due to a problem with the name server. */
#define	RC_NAME_ERROR	3	/* Name Error - Meaningful only for responses from an authoritative name server, this code signifies that the domain name referenced in the query does not exist.*/
#define	RC_NOT_IMPL		4	/* Not Implemented - The name server does not support the requested kind of query.					*/
#define	RC_REFUSED		5	/* Refused - The name server refuses to perform the specified operation for policy reasons. 		*/
 							/* For example, a name server may not wish to provide the information to the particular requester,	*/
							/* or a name server may not wish to perform a particular operation (e.g., zone 						*/

/* rd : Recursion desired , tc : Truncation, aa	: Authoratative answer, opcode : op code = OP_QUREY, OP_IQUREY, OP_STAUTS, qr : Query/Response */
#define MAKE_FLAG0(qr,op,aa,tc,rd)	( ( ( qr & 0x01 ) << 7 ) + ( ( op & 0x0F ) << 3 ) + ( ( aa & 0x01 ) << 2 ) + ( ( tc & 0x01 ) << 1 ) + ( rd & 0x01 ) )

/* rcode : Response code, z : Reserved for future use.  Must be zero in all queries and responses, */
/* ra : Recursion Available - this be is set or cleared in a response, and denotes whether recursive query support is available in the name server.*/
#define MAKE_FLAG1(ra,z,rcode)		( ( ( ra & 0x01 ) << 7 ) + ( ( z & 0x07 ) << 4 ) +  ( rcode & 0x0F ) )

/* Header for all domain messages	*/
typedef struct DHDR_t {
	uint16_t id;			/* Identification					*/
	uint8_t flag0;
	uint8_t flag1;
	uint16_t qdcount;		/* Question count					*/
	uint16_t ancount;		/* Answer count						*/
	uint16_t nscount;		/* Authority (name server) count	*/
	uint16_t arcount;		/* Additional record count			*/
} DHDR;

/* QUESTION FORMAT					*/
typedef struct QUESTION_t {
	uint16_t qtype;
	uint16_t qclass;
} DQST;

/* Resource record format			*/
typedef struct RESOURCE_RECORD_t {
	uint16_t _type;
	uint16_t _class;
	uint16_t _ttl;
	uint16_t _rdlen;
} DRR;

/* Query type						*/
typedef enum _QUERYDATA {
	BYNAME ,
	BYIP
} QUERYDATA;

#define DHDR_SIZE 12
#define	COMPRESSION_SCHEME	0xC0

//#define MAX_DNSMSG_SIZE 	512				/* Maximum size of DNS message(default 512)	*/
#define MAX_DNSMSG_SIZE 	320				/* Maximum size of DNS message(default 512)	*/
#define MAX_DOMAINNAME_LEN	50				/* Maximum size of domain name(default 50 )	*/
#define MAX_QNAME_LEN		128				/* Maximum size of qname (default 128 )		*/

/* [Jϐ錾						*/
static uint16_t dns_id = 0x1122;			/* DNS query id  			*/
static uint8_t *dns_buf;
static uint8_t *get_domain_name;
static uint32_t	get_domain_ip;				/* Resolved ip address		*/
static QUERYDATA query_data;				/* Query type 				*/

/* [J֐vg^Cv錾			*/
static uint8_t dns_query( SOCKET sok ,uint32_t dnsip ,uint8_t *domain_name ,uint32_t *domain_ip ,QUERYDATA querydata ,uint16_t elapse );
static uint16_t dns_makequery( uint8_t op ,char *qname );
static uint8_t dns_parse_reponse( void );
static uint8_t * dns_parse_question(uint8_t * cp);
static uint8_t * dns_answer(uint8_t *cp);
static uint16_t parse_name( uint8_t *cp ,char *qname ,uint16_t qname_maxlen );

#if ( USE_GETHOSTBYADDR == 1 )
/*****************************************************************************
*						IPAhXhC擾
* Fuint32_t ipaddr  IPAhX
*       char     *domain hC
* ߒlFuint8_t          0:s 1:
*****************************************************************************/
uint8_t gethostbyaddr( uint32_t ipaddr ,char *domain )
{
	uint8_t buf[ MAX_DNSMSG_SIZE + MAX_DOMAINNAME_LEN + MAX_QNAME_LEN + 16 ];
	uint32_t dns_server_ip = netInfo.ip;

	dns_buf = buf;

	if ( ( dns_server_ip == 0 ) || ( dns_server_ip == 0xFFFFFFFF ) ) {
		#if DEBUG
		puts_P( PSTR("DNS server ip address is not configured.") );
		#endif
		return ( 0 );
	}

	if ( !dns_query( DNSC_SOCK ,dns_server_ip ,(uint8_t *)domain ,&ipaddr ,BYIP ,100 ) ) {
		#if DEBUG
		printf_P( PSTR("Fail to communicate with DNS server[%s]\n") ,inet_ntoa( dns_server_ip ) );
		#endif
		return ( 0 );
	}

	return ( 1 );
}  
#endif
 
/*****************************************************************************
*						zXgIPAhX擾
* Fchar     *hostname zXg
* ߒlFuint32_t IPAhX
*****************************************************************************/
uint32_t gethostbyname( char *hostname )
{
	uint8_t buf[ MAX_DNSMSG_SIZE + MAX_DOMAINNAME_LEN + MAX_QNAME_LEN + 16 ];
	uint32_t dns_server_ip = netInfo.dns;
	uint32_t hostip;

	dns_buf = buf;

	if ( ( dns_server_ip == 0 ) || ( dns_server_ip == 0xFFFFFFFF ) ) {
		#if DEBUG
		puts_P( PSTR("DNS server ip address is not configured.") );
		#endif
		return ( 0 );
	}

	#if DEBUG
	printf_P( PSTR("DNS SERVER : %s\n") ,inet_ntoa( dns_server_ip ) );
	#endif

	hostip = inet_addr( hostname );		/* IPAhX̏ꍇ̕ϊ	*/
	if ( hostip ) {
		return ( hostip );
	}
	hostip = 0L;

	if( !dns_query( DNSC_SOCK ,dns_server_ip ,(uint8_t *)hostname ,&hostip ,BYNAME ,100 ) ) {
		#if DEBUG
		printf_P( PSTR("Fail to communicate with DNS server[%s]\n") ,inet_ntoa( dns_server_ip ) );
		#endif
		hostip = 0;
	}

	return ( hostip );
}

/*****************************************************************************
* Transfer query message for user entered domain name to desiginated DNS Server
* FSOCKET    sok          \Pbgԍ
*		uint32_t  dnsip         DNST[o[IPAhX
*		uint32_t  *domain_name hC
*		uint32_t  *domain_ip   hCIPAhX
*		QUERYDATA querydata    ( BYNAME or BYIP )
*		uint16_t  elapse       ^CAEg(10msecP)
* ߒlFuint8_t                0:s 1:
*****************************************************************************/
static uint8_t dns_query( SOCKET sok ,uint32_t dnsip ,uint8_t *domain_name ,uint32_t *domain_ip ,QUERYDATA querydata ,uint16_t elapse )
{
	uint16_t len;
	uint16_t port;
	uint8_t reponse_received = 0;
	uint8_t *qname = NULL;

	get_domain_name = dns_buf + MAX_DNSMSG_SIZE;

	query_data = querydata;

	if ( querydata == BYNAME ) {
		qname = domain_name;

	} else {
		qname = get_domain_name + MAX_QNAME_LEN;

		#if DEBUG
		printf_P( PSTR("domain_ip=%08lx\n") ,domain_ip );
		#endif

		if( *domain_ip != swapByteorder32(*domain_ip) ) {		/* if domain_ip is little-endian		*/
			inet_ntoa( (char *)qname ,*domain_ip );
		} else {												/* if domain_ip is big-endian then make the reverse ip address string. */
			inet_ntoa( (char *)qname ,swapByteorder32( *domain_ip ) );
		}
		strcat_P( (char *)qname ,PSTR(".in-addr.arpa") );
	}

	if ( udp_open( sok ,PORT_AUTO ) != SOCK_UDP ) {
		return ( 0 );
	}

	#if DEBUG
	printf_P( PSTR("Querying server %s by %s\n") ,inet_ntoa( dnsip ) ,qname );
	#endif

	len = dns_makequery( OP_QUERY ,(char *)qname );					/* create DNS query					*/
	if ( !len ) {
		#if DEBUG
		puts_P( PSTR("Fail to make query") );
		#endif		
		return ( 0 );
	}

	udp_sendto( sok ,dns_buf ,len ,dnsip ,IPPORT_DOMAIN );			/* DNS query send to DNS server		*/

	#if DEBUG
	printf_P( PSTR("sent dns query to DNS server : [%s]\n") ,inet_ntoa( dnsip ) );		
	#endif	

	while ( elapse-- > 0 ) {
		_delay_ms( 10 );											/* wait until Domain name resolution*/
		if ( ( len = wiz_getSn_RX_RSR( sok ) ) > 0 ) {
			if ( len > MAX_DNSMSG_SIZE ) {
				len = MAX_DNSMSG_SIZE;
			}
			len = udp_recvfrom( sok ,dns_buf ,len ,&dnsip ,&port );
			reponse_received = 1;
			break;			
		}
	}
	udp_close( sok );

	if( !reponse_received ) {
		#if DEBUG
		puts_P( PSTR("No response from DNS server") );
		#endif
		return ( 0 );
	}

	if( !dns_parse_reponse() ) {								/* Convert to local format				*/
		#if DEBUG
		puts_P( PSTR("Fail to parse reponse from DNS server") );
		#endif
		return ( 0 );
	}

	if ( query_data == BYNAME ) {
		*domain_ip = get_domain_ip;
	} else {
		strcpy( (char*)domain_name ,(char*)get_domain_name );
	}

	return ( 1 );
}

/*****************************************************************************
*             Create query message for transferring to DNS server
* Fuint8_t op          opcode
*		char *qname size of query message
* ߒlFuint16_t            NGTCY
*****************************************************************************/
static uint16_t dns_makequery( uint8_t op ,char *qname )
{
	uint8_t *query  = dns_buf;
	uint8_t *domain_tok;
	uint16_t domain_len;
	uint16_t qtype	= (query_data) ? TYPE_PTR : TYPE_A;	/* Domain name pointer or Host Address	*/
	uint16_t qclass	= CLASS_IN;							/* Internet								*/

	/* Make Qurey Header */
	memset( dns_buf ,0 ,MAX_DNSMSG_SIZE );
	*(uint16_t *)query = swapByteorder16( dns_id );

	#if DEBUG
	printf_P( PSTR("query : id = %02X%02X, ") ,query[0] ,query[1] );
	#endif

	query += 2;
	*query = MAKE_FLAG0( QR_QUERY ,op ,0 ,0 ,1 );		/* Recursion desired					*/

	#if DEBUG
	printf_P( PSTR("opcode = %02X, ") ,*query );
	#endif

	query++;
	*query = MAKE_FLAG1( 0 ,0 ,0 );

	#if DEBUG
	printf_P( PSTR("rcode = %02X\n") ,*query );
	#endif	

	query++;
	*(uint16_t*)query = swapByteorder16( 1 );

	query = dns_buf + DHDR_SIZE;

	/* Make Question Section */
	do {
		domain_tok = (uint8_t *)strchr( (char *)qname ,'.' );
		if ( domain_tok ) {
			domain_len = ( (uint16_t)domain_tok - (uint16_t)qname ) & 0xFF;
		} else {
			domain_len = strlen( qname );
		}

		if ( domain_len > 63 ) {
			#if DEBUG
			puts_P( PSTR("Invalid label length because labels are restricted to 63 octets or less.") );
			#endif
			return ( 0 );	/* since the label must begin with two zero bits because labels are restricted to 63 octets or less. */
		}

		*query++ = domain_len;
		memcpy( query ,qname ,domain_len );
		qname += domain_len + 1;
		query += domain_len;

	} while ( domain_tok );

	*query++ = '\0';									/* terminate QNAME field with 'NULL'	*/

	/* fill in QTYPE field, for host address	*/
	*query++ = qtype >> 8 & 0xFF;
	*query++ = qtype & 0xFF;
	
	/* fill in QCLASS field, for internet		*/
	*query++ = qclass >> 8 & 0xFF;
	*query++ = qclass & 0xFF;

	return ( (uint16_t)query - (uint16_t)dns_buf );		/* return the size of generated query	*/
}

/*****************************************************************************
*       Analyze received DNS message packet and store it into structure
* Fvoid
* ߒlFuint8_t 0Fs@1F
*****************************************************************************/
static uint8_t dns_parse_reponse( void )
{
	uint16_t i;
	DHDR dhdr;
	char *cur_ptr = (char *)dns_buf;

	dhdr.id = swapByteorder16(*((uint16_t*)cur_ptr));
	if ( dhdr.id != dns_id ) {
		#if DEBUG
		printf_P( PSTR("Responsed ID != Query ID : %d ! = %d\n") ,dhdr.id ,dns_id ); 
		#endif
		return ( 0 );
	}

	dns_id++;
	cur_ptr += 2;
	dhdr.flag0 = *cur_ptr++;
	dhdr.flag1 = *cur_ptr++;
	if( !( dhdr.flag0 & 0x80 ) || !( dhdr.flag1 & 0x80 ) ) {
		#if DEBUG
		printf_P( PSTR("No reponse message, flag0 = 0x%02X, flag1 = 0x%02X\n") ,dhdr.flag0 ,dhdr.flag1 );
		#endif
		return ( 0 );
	}

	if ( dhdr.flag1 & 0x0F ) {
		#if DEBUG
		printf_P( PSTR("Error of reponse : ") );
		uint8_t res = dhdr.flag1 & 0x0F;
		if ( res == RC_FORMAT_ERROR ) {
			puts_P( PSTR("Format Error") );
		} else if ( res == RC_SERVER_FAIL ) {
			puts_P( PSTR("Server failure") );

		} else if ( res == RC_NAME_ERROR ) {
			puts_P( PSTR("Name Error") );

		} else if ( res == RC_NOT_IMPL ) {
			puts_P( PSTR("Not Implemented") );

		} else if ( res == RC_REFUSED ) {
			puts_P( PSTR("Refused") );
		}
		#endif
		return ( 0 );
	}	

	dhdr.qdcount = swapByteorder16( *(uint16_t *)cur_ptr );
	cur_ptr += 2;

	dhdr.ancount = swapByteorder16( *(uint16_t *)cur_ptr );
	cur_ptr += 2;

	dhdr.nscount = swapByteorder16( *(uint16_t*)cur_ptr );
	cur_ptr += 2;

	dhdr.arcount = swapByteorder16( *(uint16_t*)cur_ptr );	
	cur_ptr += 2;

	#if DEBUG
	printf_P( PSTR("Response : question count = %d, answer count = %d\n") ,dhdr.qdcount ,dhdr.ancount );
	printf_P( PSTR("Response : authority count = %d, additiional count = %d\n") ,dhdr.nscount ,dhdr.arcount );
	#endif	

	/* Now parse the variable length sections */
	for ( i = 0 ;i < dhdr.qdcount ;i++ ) {
		cur_ptr = (char *)dns_parse_question( (uint8_t *)cur_ptr );	/* Question section	*/
		if( !cur_ptr ) {
			#if DEBUG
			printf_P( PSTR("Fail to parse question section%d\n") ,i );	
			#endif
			return ( 0 );
		}
	}
		
	/* parse resource records */
	for ( i = 0 ;i < dhdr.ancount ;i++ ) {
		cur_ptr = (char *)dns_answer( (uint8_t *)cur_ptr );			/* Answer section	*/
		if ( !cur_ptr ) {
			#if DEBUG
			printf_P( PSTR("Fail to parse answer section%d\n") ,i );
			#endif
			return ( 0 );
		}
	}

	return ( 1 );
}

/*****************************************************************************
*                Parse question section in the DNS query
* Fuint8_t *cp bZ[Wւ̃|C^
* ߒlFuint8_t *   ̃bZ[Wւ̃|C^
*****************************************************************************/
static uint8_t *dns_parse_question( uint8_t *cp )
{
	uint16_t len;
	char name[ MAX_QNAME_LEN ];

	len = parse_name( cp ,name ,sizeof(name) );
	if ( !len ) {
		#if DEBUG
		puts_P( PSTR("Fail to parse (Q)NAME field") );
		#endif
		return ( NULL );
	}

	cp += len;
	cp += 2;								/* skip type	*/
	cp += 2;								/* skip class	*/

	#if DEBUG
	printf_P( PSTR("In question section, (Q)NAME field value : %s") ,name );
	#endif	

	return ( cp );
}

/*****************************************************************************
*                Parse answer section in the DNS query.
* Fuint8_t *cp bZ[Wւ̃|C^
* ߒlFuint8_t *   ̃bZ[Wւ̃|C^
*****************************************************************************/
static uint8_t *dns_answer( uint8_t *cp )
{
	char qname[MAX_QNAME_LEN];
	uint16_t len, type;
	uint32_t tip;

	len = parse_name( cp, qname, sizeof(qname));
	
	if ( !len ) {
		return ( NULL );
	}

	cp += len;
	type = *cp++;
	type = ( type << 8 ) + (uint16_t)*cp++;	/* type				*/
	cp += 2;								/* skip class		*/
	cp += 4;								/* skip ttl			*/
	cp += 2;								/* skip len			*/

	switch( type ) {
		case TYPE_A:
			tip = 0;
			*(uint8_t*)&tip = *cp++;		/* Network odering	*/
			*( (uint8_t*)&tip + 1 ) = *cp++;
			*( (uint8_t*)&tip + 2 ) = *cp++;
			*( (uint8_t*)&tip + 3 ) = *cp++;

			if ( query_data == BYNAME ) {
				get_domain_ip = tip;
			}

			#if DEBUG
			printf_P( PSTR("RRs : TYPE_A = %s\n") ,inet_ntoa( tip ) );
			#endif

			break;

		case TYPE_CNAME:
		case TYPE_MB:
		case TYPE_MG:
		case TYPE_MR:
		case TYPE_NS:
		case TYPE_PTR:
			len = parse_name( cp ,qname ,sizeof(qname) );			/* These types all consist of a single domain name	*/
			if( !len ) {
				return ( 0 );										/* convert it to ascii format						*/
			}
			cp += len;
			if( ( query_data == BYIP ) && ( type == TYPE_PTR ) ) {
				strcpy( (char *)get_domain_name ,qname );

				#if DEBUG
				printf_P( PSTR("RRs : TYPE_PTR  = %s\n") ,qname );
				#endif
			}
			break;

		case TYPE_HINFO:
			len = *cp++;
			cp += len;

			len = *cp++;
			cp += len;
			break;

		case TYPE_MX:
			cp += 2;
			len = parse_name( cp ,qname ,sizeof(qname) );		/* Get domain name of exchanger						*/
			if ( !len ) {
				#if DEBUG
				puts_P( PSTR("TYPE_MX : Fail to get domain name of exechanger") );
				#endif
				return ( 0 );
			}
			cp += len;
			break;

		case TYPE_SOA:
			len = parse_name( cp ,qname ,sizeof(qname) );			/* Get domain name of name server				*/
			if ( !len ) {
				#if DEBUG
				puts_P( PSTR("TYPE_SOA : Fail to get domain name of name server") );
				#endif
				return ( 0 );
			}
			cp += len;

			len = parse_name( cp ,qname ,sizeof(qname) );		/* Get domain name of responsible person			*/
			if( !len ) {
				#if DEBUG
				puts_P( PSTR("TYPE_SOA : Fail to get domain name of responsible person") );
				#endif
				return ( 0 );
			}
			cp += len;

			cp += 4;
			cp += 4;
			cp += 4;
			cp += 4;
			cp += 4;
			break;

		case TYPE_TXT:
			break;						/* Just stash	*/

		default:
			break;						/* Ignore		*/
	}

	return ( cp );
}

/*****************************************************************************
*				Parse answer section in the DNS query.
* Fuint8_t  *cp          bZ[Wւ̃|C^
*       char     *qname       擾i[̈
*       uint16_t qname_maxlen qnamẽTCY
* ߒlFuint16_t              擾f[^TCY
****************************************************************************/
static uint16_t parse_name( uint8_t *cp ,char *qname ,uint16_t qname_maxlen )
{
	uint16_t slen;				/* Length of current segment			*/
	uint16_t clen = 0;			/* Total length of compressed name		*/
	uint16_t indirect = 0;		/* Set if indirection encountered		*/
	uint16_t nseg = 0;			/* Total number of label in name		*/

	while ( 1 ) {
		slen = *cp++;			/* Length of this segment				*/
		if ( !indirect ) {
			clen++;
		}

		if ( ( slen & 0xc0 ) == 0xc0 ) {	/* Is used in compression scheme?	*/
			cp = &dns_buf[ ( ( slen & 0x3f ) << 8 ) + *cp ];	/* Follow indirection	*/
			if( !indirect )	{
				clen++;
			}
			indirect = 1;			
			slen = *cp++;
		}

		if ( slen == 0 ) {		/* zero length == all done				*/
			break;
		}

		if ( !indirect ) {
			clen += slen;
		}

		if ( ( qname_maxlen -= slen + 1 ) < 0 ) {
			#if DEBUG
			puts_P( PSTR("Not enough memory") );
			#endif
			return ( 0 );
		}

		while ( slen-- != 0 ) {
			*qname++ = (char)*cp++;
		}
		*qname++ = '.';

		nseg++;
	}

	if ( nseg == 0 ) {
		*qname++ = '.';			/* Root name; represent as single dot	*/
	} else {
		qname--;
	}
	*qname = '\0';

	#if DEBUG
	printf_P( PSTR("Result of parsing (Q)NAME field : %s\n") ,qname );
	#endif	

	return ( clen );			/* Length of compressed message			*/
}
